package filterlist

import "github.com/AdguardTeam/urlfilter/rules"

// RuleStorageScanner scans multiple RuleScanner instances
// The rule index is built from the rule index in the list + the list ID
// First 4 bytes is the rule index in the list
// Second 4 bytes is the list ID
type RuleStorageScanner struct {
	// Scanners is the list of list scanners backing this combined scanner
	Scanners []*RuleScanner

	currentScanner    *RuleScanner
	currentScannerIdx int // Index of the current scanner
}

// Scan advances the RuleStorageScanner to the next rule, which will then be available
// through the Rule method. It returns false when the scan stops, either by
// reaching the end of the input or an error.
func (s *RuleStorageScanner) Scan() bool {
	if len(s.Scanners) == 0 {
		return false
	}

	if s.currentScanner == nil {
		s.currentScannerIdx = 0
		s.currentScanner = s.Scanners[s.currentScannerIdx]
	}

	for {
		scan := s.currentScanner.Scan()
		if scan {
			return true
		}

		// Take the next scanner or just return false if there's nothing more
		if s.currentScannerIdx == (len(s.Scanners) - 1) {
			return false
		}

		s.currentScannerIdx++
		s.currentScanner = s.Scanners[s.currentScannerIdx]
	}
}

// Rule returns the most recent rule generated by a call to Scan, and the index of this rule.
// See ruleListIdxToStorageIdx for more information on what this index is.
func (s *RuleStorageScanner) Rule() (rules.Rule, int64) {
	if s.currentScanner == nil {
		return nil, 0
	}

	f, idx := s.currentScanner.Rule()
	if f == nil {
		return nil, 0
	}

	return f, ruleListIdxToStorageIdx(int32(f.GetFilterListID()), int32(idx))
}

// ruleListIdxToStorageIdx converts pair of listID and rule list index
// to a single int64 "storage index"
func ruleListIdxToStorageIdx(listID int32, ruleIdx int32) int64 {
	return int64(listID)<<32 | int64(ruleIdx)&0xFFFFFFFF
}

// storageIdxToRuleListIdx converts the "storage index" to two integers:
// listID -- rule list identifier
// ruleIdx -- index of the rule in the list
func storageIdxToRuleListIdx(storageIdx int64) (listID int32, ruleIdx int32) {
	listID = int32(storageIdx >> 32)
	ruleIdx = int32(storageIdx)
	return
}
